==> spirit_hub-0.0.1.10.tar.gz: spirit_hub-0.0.1.10/app/services/spirit_hub/operation_tree_model.rb <==
# frozen_string_literal: true

require 'csv'
module SpiritHub
  class OperationTreeModel
    BLOCKS_NAMES = ['Dépôt', 'Fichiers valides', 'Fichiers non valides'].freeze
    CSV_HEADERS = [
      'Code',
      'Libelle',
      'Code parent',
      'Lecture',
      'Ecriture',
      'Autre instance',
      'Phase',
      'Phase liee',
      'Nombre de mois',
      'Chapitre',
      'Sous chapitre',
      'Type de document'
    ].freeze

    class << self
      %i[
        create_batch
        validate_models
        find_valid_docs
        generate_operation_tree
      ].each do |method_name|
        define_method(method_name) do |*args|
          new(*args).public_send(method_name)
        end
      end
    end

    def initialize(operation = nil)
      @operation = operation
      @batch = Batch.find_by(name: 'Classification plans models')
      @tagset_batch_id = @batch&.tagset_id
      @tagset_batch = Tagset.find_by(id: @tagset_batch_id)
      @deposit_category_id = @tagset_batch && Category.find_by(name: 'Dépôt', cycle_id: @tagset_batch.blocset_id)&.id
      @valid_doc_category_id = @tagset_batch && Category.find_by(name: 'Fichiers valides', cycle_id: @tagset_batch.blocset_id)&.id
      @invalid_doc_category_id = @tagset_batch && Category.find_by(name: 'Fichiers non valides', cycle_id: @tagset_batch.blocset_id)&.id
    end

    def create_batch
      @batch ||= Batch.create(name: 'Classification plans models')
      @batch.update(writers: [admin.id]) if admin
      @tagset_batch_id ||= @batch.create_tagset
      @tagset_batch ||= Tagset.find_by(id: @tagset_batch_id)

      @tagset_batch.update(cycle: nil)
      @tagset_batch.update(writers: [admin.id]) if admin

      BLOCKS_NAMES.each do |bloc_name|
        Category.find_or_create_by(name: bloc_name, cycle_id: @tagset_batch.blocset_id)
      end

      @batch.update(inline_script: "SpiritHub::OperationTreeModel.validate_models\n")
    end

    def validate_models
      Category.find(@deposit_category_id).doc_ids(@tagset_batch_id).each do |doc_id|
        doc = Doc.find(doc_id)
        @operation_name = doc.title.gsub('.csv', '')
        extract_values(doc)
        root_tagset = create_tree
        @tagset_batch.update(tagset_ids: @tagset_batch.tagset_ids.to_a + [root_tagset.id])
        doc.set_category(@tagset_batch_id, @valid_doc_category_id)
        puts 'Tree created successfully'
        puts "Created tagsets : #{@created_tagsets.length}"
        puts "Created completeness : #{@created_completeness.length}"
      rescue StandardError => e
        puts "Error: #{e.message}"
        puts 'Backtrace:'
        puts e.backtrace.join("\n")
        doc.set_category(@tagset_batch_id, @invalid_doc_category_id)
      end
    end

    def add_doc_to_valid_category(doc)
      doc.set_category(@tagset_batch_id, @valid_doc_category_id)
    end

    def find_valid_docs
      return [] unless (valid_doc_category = Category.find_by(id: @valid_doc_category_id))

      Doc.where(:id.in => valid_doc_category.doc_ids(@tagset_batch_id).to_a)
    end

    def generate_operation_tree
      @operation_name = @operation.title
      extract_values(@operation.classement_plan)
      root_tagset = create_tree
      @operation.set(root_tagset_id: root_tagset.id, readers: root_tagset.readers, writers: root_tagset.writers)
    end

    private

    def admin
      @admin ||= User.find_by(email: 'support@plugandwork.fr') || User.find_by(roles: 'Administrator')
    end

    def extract_values(doc)
      puts "Trying to create tree for doc #{doc.title}"
      @tagset_values = []
      @completeness_values = []
      @phase_values = []
      @group_names = []

      entry_file_table = CSV.parse(doc.file.data, col_sep: ';', headers: true)
      headers = entry_file_table.first.to_h.keys.map

      raise "Invalid CSV headers for doc #{doc.title} : #{headers} not match #{CSV_HEADERS}" unless headers.to_set == CSV_HEADERS.to_set

      entry_file_table.each do |row|
        check_dependencies(row, entry_file_table)
        extract_values_row(row)
      end
    end

    def check_dependencies(row, entry_file_table)
      if !row['Code parent'].blank? && entry_file_table.find { |r| r['Code'] == row['Code parent'] }.nil?
        raise "Invalid parent code for #{row['Libelle']} : #{row['Code parent']} not found"
      end

      return unless !row['Phase liee'].blank? && entry_file_table.find { |r| r['Phase'] == row['Phase liee'] }.nil?

      raise "Invalid linked phase for #{row['Libelle']} : #{row['Phase liee']} not found"
    end

    def extract_values_row(row) # rubocop:disable Metrics/AbcSize
      title = clean_string(row['Libelle'])
      code = row['Code'].blank? ? nil : clean_string(row['Code'])
      parent_code = clean_string(row['Code parent'])
      reader_group_names = extract_group_names(row['Lecture'])
      writer_group_names = extract_group_names(row['Ecriture'])
      max_duplicates = clean_string(row['Autre instance'])
      @group_names = (@group_names + reader_group_names.to_a + writer_group_names.to_a).uniq
      phase_name = nil

      if (phase_value = extract_phase_value(row))
        @phase_values << phase_value unless @phase_values.find { |v| v[:title] == phase_value[:title] }
        phase_name = phase_value[:title]
      end

      superfields = extract_superfields(row)

      if row['Code'] == 'DOCUMENT'
        @completeness_values << { title:, parent_code:, superfields: }
      else
        @tagset_values << { title:, code:, parent_code:, reader_group_names:, writer_group_names:, phase_name:, max_duplicates:, superfields: }
      end
    end

    def clean_string(string)
      string&.strip&.force_encoding('UTF-8')
    end

    def extract_group_names(string)
      return [] if string.blank?

      string&.split(',')&.map { |g| clean_string(g) }
    end

    def extract_phase_value(row)
      return if row['Phase'].blank?

      { title: clean_string(row['Phase']), parent_name: clean_string(row['Phase liee']), number_of_months: row['Nombre de mois'] }
    end

    def extract_superfields(row)
      superfields = {}
      superfields[:chapter] = clean_string(row['Chapitre']) unless row['Chapitre'].blank?
      superfields[:sub_chapter] = clean_string(row['Sous chapitre']) unless row['Sous chapitre'].blank?
      superfields[:document_type] = remove_leading_and_trailing_zeros(clean_string(row['Type de document'])) unless row['Type de document'].blank?
      superfields[:code] = clean_string(row['Code'])
      superfields
    end

    def remove_leading_and_trailing_zeros(value)
      value.sub(/^0+/, '').sub(/0+$/, '')
    end

    def create_tree
      @created_tagsets = []
      @created_completeness = []
      @created_phases = []

      @group_names.each do |group_name|
        Group.find_or_create_by(name: group_name)
      end

      @phase_values.each do |value|
        create_phase(value)
      end

      root_tagset = create_tagset(title: "Projet #{@operation_name}", code: nil)
      @created_tagsets << root_tagset

      @completeness_values.each do |value|
        create_completeness(value)
      end

      root_tagset
    end

    def create_tagset(value) # rubocop:disable Metrics/AbcSize,Metrics/PerceivedComplexity
      tagset_ids = []
      readers = Group.where(:name.in => value[:reader_group_names].to_a).pluck(:id).map { |id| "g_#{id}" }
      writers = Group.where(:name.in => value[:writer_group_names].to_a).pluck(:id).map { |id| "g_#{id}" }

      @tagset_values.select { |v| v[:parent_code] == value[:code] }.each do |children_value|
        childen_tagset = create_tagset(children_value)
        @created_tagsets << childen_tagset
        tagset_ids << childen_tagset.id
        readers += childen_tagset.readers.to_a
        writers += childen_tagset.writers.to_a
      end

      tagset_data = {
        is_space: true,
        title: value[:title],
        tagset_ids:,
        readers:,
        writers:,
        superfields: value[:superfields],
        responsible: admin&.id
      }

      if @operation
        tagset_data.merge!(operation_tagset_data(value))
      else
        tagset_data.merge!(writers: writers + [admin&.id.to_s])
      end

      Tagset.create(tagset_data)
    end

    def operation_tagset_data(value)
      phase = @created_phases.find { |p| p.title == value[:phase_name] }

      {
        operation_id: @operation.id,
        type: 'directory',
        responsible: @operation.responsible,
        current_user_id: @operation.current_user_id,
        start_date: @operation.start_date,
        end_date: @operation.end_date,
        spirit_hubdoc: true,
        max_duplicates: value[:max_duplicates],
        phase_id: phase&.id
      }
    end

    def create_completeness(value)
      parent_tagset = @created_tagsets.find { |t| t.superfields[:code] == value[:parent_code] }
      completeness = if @operation
                       SpiritHub::Completeness.create(title: value[:title], tagset_id: parent_tagset.id, superfields: value[:superfields])
                     else
                       Doc.create(title: value[:title], tagset_ids: [parent_tagset.id], superfields: value[:superfields])
                     end
      @created_completeness << completeness
    end

    def create_phase(value)
      return unless @operation

      parent_phase = nil
      if value[:parent_name]
        parent_phase_value = @phase_values.find { |v| v[:title] == value[:parent_name] }
        parent_phase = @created_phases.find { |p| p.title == parent_phase_value[:title] }
        parent_phase ||= create_phase(parent_phase_value)
      end

      initial_date = @operation.start_date + value[:number_of_months].to_i.months
      phase = Phase.find_or_create_by(title: value[:title], operation_id: @operation.id, initial_date:)
      phase.update(parent_id: parent_phase&.id)
      @created_phases << phase
      phase
    end
  end
end

==> spirit_hub-0.0.1.11.tar.gz: spirit_hub-0.0.1.11/app/services/spirit_hub/operation_tree_model.rb <==
# frozen_string_literal: true

require 'csv'
module SpiritHub
  class OperationTreeModel
    BLOCKS_NAMES = ['Dépôt', 'Fichiers valides', 'Fichiers non valides'].freeze
    CSV_HEADERS = [
      'Code',
      'Libelle',
      'Code parent',
      'Lecture',
      'Ecriture',
      'Autre instance',
      'Phase',
      'Phase liee',
      'Nombre de mois',
      'Chapitre',
      'Sous chapitre',
      'Type de document'
    ].freeze

    class << self
      %i[
        create_batch
        validate_models
        find_valid_docs
        generate_operation_tree
      ].each do |method_name|
        define_method(method_name) do |*args|
          new(*args).public_send(method_name)
        end
      end
    end

    def initialize(operation = nil, user = nil)
      @operation = operation
      @user = user
      @batch = Batch.find_by(name: 'Classification plans models')
      @tagset_batch_id = @batch&.tagset_id
      @tagset_batch = Tagset.find_by(id: @tagset_batch_id)
      @deposit_category_id = @tagset_batch && Category.find_by(name: 'Dépôt', cycle_id: @tagset_batch.blocset_id)&.id
      @valid_doc_category_id = @tagset_batch && Category.find_by(name: 'Fichiers valides', cycle_id: @tagset_batch.blocset_id)&.id
      @invalid_doc_category_id = @tagset_batch && Category.find_by(name: 'Fichiers non valides', cycle_id: @tagset_batch.blocset_id)&.id
    end

    def create_batch
      @batch ||= Batch.create(name: 'Classification plans models')
      @batch.update(writers: [admin.id]) if admin
      @tagset_batch_id ||= @batch.create_tagset
      @tagset_batch ||= Tagset.find_by(id: @tagset_batch_id)

      @tagset_batch.update(cycle: nil)
      @tagset_batch.update(writers: [admin.id]) if admin

      BLOCKS_NAMES.each do |bloc_name|
        Category.find_or_create_by(name: bloc_name, cycle_id: @tagset_batch.blocset_id)
      end

      @batch.update(inline_script: "SpiritHub::OperationTreeModel.validate_models\n")
    end

    def validate_models
      Category.find(@deposit_category_id).doc_ids(@tagset_batch_id).each do |doc_id|
        doc = Doc.find(doc_id)
        @operation_name = doc.title.gsub('.csv', '')
        extract_values(doc)
        root_tagset = create_tree
        @tagset_batch.update(tagset_ids: @tagset_batch.tagset_ids.to_a + [root_tagset.id])
        doc.set_category(@tagset_batch_id, @valid_doc_category_id)
        puts 'Tree created successfully'
        puts "Created tagsets : #{@created_tagsets.length}"
        puts "Created completeness : #{@created_completeness.length}"
      rescue StandardError => e
        puts "Error: #{e.message}"
        puts 'Backtrace:'
        puts e.backtrace.join("\n")
        doc.set_category(@tagset_batch_id, @invalid_doc_category_id)
      end
    end

    def add_doc_to_valid_category(doc)
      doc.set_category(@tagset_batch_id, @valid_doc_category_id)
    end

    def find_valid_docs
      return [] unless (valid_doc_category = Category.find_by(id: @valid_doc_category_id))

      Doc.where(:id.in => valid_doc_category.doc_ids(@tagset_batch_id).to_a)
    end

    def generate_operation_tree
      @operation_name = @operation.title
      extract_values(@operation.classement_plan)
      root_tagset = create_tree
      @operation.set(root_tagset_id: root_tagset.id, readers: root_tagset.readers, writers: root_tagset.writers)
      SpiritHub::JournalEntry.create_action_log('OPERATION_CREATED', tagset: root_tagset, user: @user, title: @operation.title)
      root_tagset
    end

    private

    def admin
      @admin ||= User.find_by(email: 'support@plugandwork.fr') || User.find_by(roles: 'Administrator')
    end

    def extract_values(doc)
      puts "Trying to create tree for doc #{doc.title}"
      @tagset_values = []
      @completeness_values = []
      @phase_values = []
      @group_names = []

      entry_file_table = CSV.parse(doc.file.data, col_sep: ';', headers: true)
      headers = entry_file_table.first.to_h.keys.map

      raise "Invalid CSV headers for doc #{doc.title} : #{headers} not match #{CSV_HEADERS}" unless headers.to_set == CSV_HEADERS.to_set

      entry_file_table.each do |row|
        check_dependencies(row, entry_file_table)
        extract_values_row(row)
      end
    end

    def check_dependencies(row, entry_file_table)
      if !row['Code parent'].blank? && entry_file_table.find { |r| r['Code'] == row['Code parent'] }.nil?
        raise "Invalid parent code for #{row['Libelle']} : #{row['Code parent']} not found"
      end

      return unless !row['Phase liee'].blank? && entry_file_table.find { |r| r['Phase'] == row['Phase liee'] }.nil?

      raise "Invalid linked phase for #{row['Libelle']} : #{row['Phase liee']} not found"
    end

    def extract_values_row(row) # rubocop:disable Metrics/AbcSize
      title = clean_string(row['Libelle'])
      code = row['Code'].blank? ? nil : clean_string(row['Code'])
      parent_code = clean_string(row['Code parent'])
      reader_group_names = extract_group_names(row['Lecture'])
      writer_group_names = extract_group_names(row['Ecriture'])
      max_duplicates = clean_string(row['Autre instance'])
      @group_names = (@group_names + reader_group_names.to_a + writer_group_names.to_a).uniq
      phase_name = nil

      if (phase_value = extract_phase_value(row))
        @phase_values << phase_value unless @phase_values.find { |v| v[:title] == phase_value[:title] }
        phase_name = phase_value[:title]
      end

      superfields = extract_superfields(row)

      if row['Code'] == 'DOCUMENT'
        @completeness_values << { title:, parent_code:, superfields: }
      else
        @tagset_values << { title:, code:, parent_code:, reader_group_names:, writer_group_names:, phase_name:, max_duplicates:, superfields: }
      end
    end

    def clean_string(string)
      string&.strip&.force_encoding('UTF-8')
    end

    def extract_group_names(string)
      return [] if string.blank?

      string&.split(',')&.map { |g| clean_string(g) }
    end

    def extract_phase_value(row)
      return if row['Phase'].blank?

      { title: clean_string(row['Phase']), parent_name: clean_string(row['Phase liee']), number_of_months: row['Nombre de mois'] }
    end

    def extract_superfields(row)
      superfields = {}
      superfields[:chapter] = clean_string(row['Chapitre']) unless row['Chapitre'].blank?
      superfields[:sub_chapter] = clean_string(row['Sous chapitre']) unless row['Sous chapitre'].blank?
      superfields[:document_type] = remove_leading_and_trailing_zeros(clean_string(row['Type de document'])) unless row['Type de document'].blank?
      superfields[:code] = clean_string(row['Code'])
      superfields
    end

    def remove_leading_and_trailing_zeros(value)
      value.sub(/^0+/, '').sub(/0+$/, '')
    end

    def create_tree
      @created_tagsets = []
      @created_completeness = []
      @created_phases = []

      @group_names.each do |group_name|
        Group.find_or_create_by(name: group_name)
      end

      @phase_values.each do |value|
        create_phase(value)
      end

      root_tagset = create_tagset(title: "Projet #{@operation_name}", code: nil)
      @created_tagsets << root_tagset

      @completeness_values.each do |value|
        create_completeness(value)
      end

      root_tagset
    end

    def create_tagset(value) # rubocop:disable Metrics/AbcSize,Metrics/PerceivedComplexity
      tagset_ids = []
      readers = Group.where(:name.in => value[:reader_group_names].to_a).pluck(:id).map { |id| "g_#{id}" }
      writers = Group.where(:name.in => value[:writer_group_names].to_a).pluck(:id).map { |id| "g_#{id}" }

      @tagset_values.select { |v| v[:parent_code] == value[:code] }.each do |children_value|
        childen_tagset = create_tagset(children_value)
        @created_tagsets << childen_tagset
        tagset_ids << childen_tagset.id
        readers += childen_tagset.readers.to_a
        writers += childen_tagset.writers.to_a
      end

      tagset_data = {
        is_space: true,
        title: value[:title],
        tagset_ids:,
        readers:,
        writers:,
        superfields: value[:superfields],
        responsible: admin&.id
      }

      if @operation
        tagset_data.merge!(operation_tagset_data(value))
      else
        tagset_data.merge!(writers: writers + [admin&.id.to_s])
      end

      tagset = Tagset.create(tagset_data)
      SpiritHub::JournalEntry.create_action_log('DIRECTORY_CREATED', tagset:, user: @user)

      tagset
    end

    def operation_tagset_data(value)
      phase = @created_phases.find { |p| p.title == value[:phase_name] }

      {
        operation_id: @operation.id,
        type: 'directory',
        responsible: @operation.responsible,
        current_user_id: @operation.current_user_id,
        start_date: @operation.start_date,
        end_date: @operation.end_date,
        spirit_hubdoc: true,
        max_duplicates: value[:max_duplicates],
        phase_id: phase&.id
      }
    end

    def create_completeness(value)
      parent_tagset = @created_tagsets.find { |t| t.superfields[:code] == value[:parent_code] }
      completeness = if @operation
                       SpiritHub::Completeness.create(title: value[:title], tagset_id: parent_tagset.id, superfields: value[:superfields])
                     else
                       Doc.create(title: value[:title], tagset_ids: [parent_tagset.id], superfields: value[:superfields])
                     end
      @created_completeness << completeness
    end

    def create_phase(value)
      return unless @operation

      parent_phase = nil
      if value[:parent_name]
        parent_phase_value = @phase_values.find { |v| v[:title] == value[:parent_name] }
        parent_phase = @created_phases.find { |p| p.title == parent_phase_value[:title] }
        parent_phase ||= create_phase(parent_phase_value)
      end

      initial_date = @operation.start_date + value[:number_of_months].to_i.months
      phase = Phase.find_or_create_by(title: value[:title], operation_id: @operation.id, initial_date:)
      phase.update(parent_id: parent_phase&.id)
      @created_phases << phase
      phase
    end
  end
end
